@@ -78,6 +78,79 @@ foo(Baz) :- baz.
     (should (not (sweeprolog-beginning-of-next-top-term)))
     (should (= (point) 19))))
 
+(ert-deftest help-echo-for-dependency ()
+  "Test that the `help-echo' property is set correctly."
+  (let ((temp (make-temp-file "sweeprolog-help-echo-text"
+                              nil
+                              "pl"
+                              "
+:- use_module(library(lists)).
+
+foo(Foo, Bar) :- flatten(Bar, Baz), member(Foo, Baz).
+")))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (goto-char 24)
+    (should (string-match "Dependency on .*, resolves calls to flatten/2, member/2"
+                          (help-at-pt-kbd-string)))))
+
+(ert-deftest terms-at-point ()
+  "Test `sweeprolog-term-search'."
+  (let ((temp (make-temp-file "sweeprolog-terms-at-point-test"
+                              nil
+                              "pl"
+                              "
+recursive(Var) :-
+    (   true
+    ->  recursive(Bar)
+    ;   var(Baz)
+    *-> Bar is foo
+    ).
+")))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (should (equal (sweeprolog-terms-at-point 81)
+                '("Bar"
+                  "Bar is foo"
+                  "var(Baz)
+    *-> Bar is foo" "true
+    ->  recursive(Bar)
+    ;   var(Baz)
+    *-> Bar is foo"
+    "recursive(Var) :-
+    (   true
+    ->  recursive(Bar)
+    ;   var(Baz)
+    *-> Bar is foo
+    )")))))
+
+(ert-deftest predicate-location ()
+  "Test `sweeprolog-predicate-location'."
+  (should (sweeprolog-predicate-location "memory_file:new_memory_file/1")))
+
+(ert-deftest term-search ()
+  "Test `sweeprolog-term-search'."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "
+bar(bar(bar), bar{bar:bar}, [bar,bar|bar]).
+foo([Bar|Baz]).
+")))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (goto-char (point-min))
+    (sweeprolog-term-search "bar")
+    (should (= (point) 10))
+    (sweeprolog-term-search "bar")
+    (should (= (point) 24))
+    (sweeprolog-term-search "bar")
+    (should (= (point) 31))
+    (sweeprolog-term-search "bar")
+    (should (= (point) 35))
+    (sweeprolog-term-search "bar")
+    (should (= (point) 39))))
+
 (ert-deftest beginning-of-next-top-term-header ()
   "Test finding the beginning of the first top term."
   (let ((temp (make-temp-file "sweeprolog-test"
@@ -163,7 +236,6 @@ foo(Foo) :- bar.
     (should (equal (get-text-property (+ (point-min) 16)
                                       'font-lock-face)
                    '(sweeprolog-local-default-face
-                     sweeprolog-predicate-indicator-default-face
                      sweeprolog-body-default-face)))
     (should (equal (get-text-property (+ (point-min) 23)
                                       'font-lock-face)
@@ -181,6 +253,40 @@ foo(Foo) :- bar.
                    '(sweeprolog-undefined-default-face
                      sweeprolog-body-default-face)))))
 
+(ert-deftest yank-hole ()
+  "Test killing and yanking a hole as a plain variable."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "")))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (sweeprolog-insert-term-with-holes ":-" 2)
+    (should (get-text-property (point-min) 'sweeprolog-hole))
+    (call-interactively #'kill-region)
+    (call-interactively #'yank)
+    (should (not (get-text-property (point-min) 'sweeprolog-hole)))))
+
+(ert-deftest insert-term-with-holes ()
+  "Test `sweeprolog-insert-term-with-holes'."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "")))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (sweeprolog-insert-term-with-holes ":-" 2)
+    (call-interactively #'kill-region)
+    (sweeprolog-insert-term-with-holes "foo" 3)
+    (call-interactively #'kill-region)
+    (sweeprolog-insert-term-with-holes "bar" 0)
+    (call-interactively #'kill-region)
+    (sweeprolog-insert-term-with-holes ";" 2)
+    (call-interactively #'kill-region)
+    (sweeprolog-insert-term-with-holes "->" 2)
+    (should (string= (buffer-string)
+                     "foo(bar, (_->_;_), _):-_."))))
+
 (ert-deftest plunit-testset-skeleton ()
   "Tests inserting PlUnit test-set blocks."
   (let ((temp (make-temp-file "sweeprolog-test"
@@ -612,6 +718,30 @@ spam:foo --> Body.
 "
                      ))))
 
+(ert-deftest dwim-next-clause-args ()
+  "Tests inserting new clause with arguments."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "
+%!  foo(+Bar) is det.
+
+foo(bar) :- bar.
+"
+                              )))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (goto-char (point-max))
+    (sweeprolog-insert-term-dwim)
+    (should (string= (buffer-string)
+                              "
+%!  foo(+Bar) is det.
+
+foo(bar) :- bar.
+foo(Bar) :- Body.
+
+"))))
+
 (ert-deftest dwim-next-clause-module-qualified ()
   "Tests inserting new module-qualified clause."
   (let ((temp (make-temp-file "sweeprolog-test"
