@@ -4,7 +4,7 @@
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
     Copyright (c)  1985-2022, University of Amsterdam
-                              VU University Amsterdam
+			      VU University Amsterdam
 			      CWI, Amsterdam
 			      SWI-Prolog Solutions b.v.
     All rights reserved.
@@ -81,6 +81,49 @@
 #define newChoice(type, fr) LDFUNC(newChoice, type, fr)
 static Choice	newChoice(DECL_LD choice_type type, LocalFrame fr);
 
+typedef foreign_t (*Func0)(void);
+typedef foreign_t (*Func1)(term_t a1);
+typedef foreign_t (*Func2)(term_t a1, term_t a2);
+typedef foreign_t (*Func3)(term_t a1, term_t a2, term_t a3);
+typedef foreign_t (*Func4)(term_t a1, term_t a2, term_t a3, term_t a4);
+typedef foreign_t (*Func5)(term_t a1, term_t a2, term_t a3, term_t a4,
+			   term_t a5);
+typedef foreign_t (*Func6)(term_t a1, term_t a2, term_t a3, term_t a4,
+			   term_t a5, term_t a6);
+typedef foreign_t (*Func7)(term_t a1, term_t a2, term_t a3, term_t a4,
+			   term_t a5, term_t a6, term_t a7);
+typedef foreign_t (*Func8)(term_t a1, term_t a2, term_t a3, term_t a4,
+			   term_t a5, term_t a6, term_t a7, term_t a8);
+typedef foreign_t (*Func9)(term_t a1, term_t a2, term_t a3, term_t a4,
+			   term_t a5, term_t a6, term_t a7, term_t a8,
+			   term_t a9);
+typedef foreign_t (*Func10)(term_t a1, term_t a2, term_t a3, term_t a4,
+			    term_t a5, term_t a6, term_t a7, term_t a8,
+			    term_t a9, term_t a10);
+
+typedef foreign_t (*NdetFunc0)(control_t);
+typedef foreign_t (*NdetFunc1)(term_t a1, control_t);
+typedef foreign_t (*NdetFunc2)(term_t a1, term_t a2, control_t);
+typedef foreign_t (*NdetFunc3)(term_t a1, term_t a2, term_t a3, control_t);
+typedef foreign_t (*NdetFunc4)(term_t a1, term_t a2, term_t a3, term_t a4,
+			       control_t);
+typedef foreign_t (*NdetFunc5)(term_t a1, term_t a2, term_t a3, term_t a4,
+			       term_t a5, control_t);
+typedef foreign_t (*NdetFunc6)(term_t a1, term_t a2, term_t a3, term_t a4,
+			       term_t a5, term_t a6, control_t);
+typedef foreign_t (*NdetFunc7)(term_t a1, term_t a2, term_t a3, term_t a4,
+			       term_t a5, term_t a6, term_t a7, control_t);
+typedef foreign_t (*NdetFunc8)(term_t a1, term_t a2, term_t a3, term_t a4,
+			       term_t a5, term_t a6, term_t a7, term_t a8,
+			       control_t);
+typedef foreign_t (*NdetFunc9)(term_t a1, term_t a2, term_t a3, term_t a4,
+			       term_t a5, term_t a6, term_t a7, term_t a8,
+			       term_t a9, control_t);
+typedef foreign_t (*NdetFunc10)(term_t a1, term_t a2, term_t a3, term_t a4,
+				term_t a5, term_t a6, term_t a7, term_t a8,
+				term_t a9, term_t a10, control_t);
+
+
 #if COUNTING
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -105,7 +148,7 @@ count(code c, Code PC)
 { const code_info *info = &codeTable[c];
 
   counting[c].times++;
-  switch(info->argtype)
+  switch(VM_ARGTYPES(info))
   { case CA1_VAR:
     case CA1_FVAR:
     case CA1_CHP:
@@ -263,7 +306,7 @@ in several virtual machine instructions.  Currently covers:
 
 	* Pending signals
 	* pthread_cancel() requested
-        * Activation of the profiler
+	* Activation of the profiler
 	* out-of-stack signalled
 	* active depth-limit
 	* attributed variable wakeup
@@ -275,7 +318,7 @@ updateAlerted(PL_local_data_t *ld)
 { int mask = 0;
 
   WITH_LD(ld)
-  { if ( is_signalled() )		        mask |= ALERT_SIGNAL;
+  { if ( is_signalled() )			mask |= ALERT_SIGNAL;
     if ( !truePrologFlag(PLFLAG_VMI_BUILTIN) ||
 	 ld->prolog_flag.occurs_check != OCCURS_CHECK_FALSE )
       ld->slow_unify = TRUE;  /* see VMI B_UNIFY_VAR */
@@ -606,38 +649,38 @@ ssu_or_det_failed(DECL_LD LocalFrame fr)
 #define CALL_FCUTTED(argc, f, c) \
   { switch(argc) \
     { case 0: \
-	f(c); \
-        break; \
+	(*(NdetFunc0)f)(c); \
+	break; \
       case 1: \
-	f(0,(c)); \
+	(*(NdetFunc1)f)(0,(c)); \
 	break; \
       case 2: \
-	f(0,0,(c)); \
-        break; \
+	(*(NdetFunc2)f)(0,0,(c)); \
+	break; \
       case 3: \
-	f(0,0,0,(c)); \
-        break; \
+	(*(NdetFunc3)f)(0,0,0,(c)); \
+	break; \
       case 4: \
-	f(0,0,0,0,(c)); \
-        break; \
+	(*(NdetFunc4)f)(0,0,0,0,(c)); \
+	break; \
       case 5: \
-	f(0,0,0,0,0,(c)); \
-        break; \
+	(*(NdetFunc5)f)(0,0,0,0,0,(c)); \
+	break; \
       case 6: \
-	f(0,0,0,0,0,0,(c)); \
-        break; \
+	(*(NdetFunc6)f)(0,0,0,0,0,0,(c)); \
+	break; \
       case 7: \
-	f(0,0,0,0,0,0,0,(c)); \
-        break; \
+	(*(NdetFunc7)f)(0,0,0,0,0,0,0,(c)); \
+	break; \
       case 8: \
-	f(0,0,0,0,0,0,0,0,(c)); \
-        break; \
+	(*(NdetFunc8)f)(0,0,0,0,0,0,0,0,(c)); \
+	break; \
       case 9: \
-	f(0,0,0,0,0,0,0,0,0,(c)); \
-        break; \
+	(*(NdetFunc9)f)(0,0,0,0,0,0,0,0,0,(c)); \
+	break; \
       case 10: \
-	f(0,0,0,0,0,0,0,0,0,0,(c)); \
-        break; \
+	(*(NdetFunc10)f)(0,0,0,0,0,0,0,0,0,0,(c)); \
+	break; \
       default: \
 	assert(0); \
     } \
@@ -671,7 +714,8 @@ discardForeignFrame(DECL_LD LocalFrame fr)
 
   fid = PL_open_foreign_frame();
   if ( true(def, P_VARARG) )
-  { (*function)(0, argc, &context);
+  { typedef foreign_t (*FuncN)(term_t av, int argc, control_t);
+    (*(FuncN)function)(0, argc, &context);
   } else
   { CALL_FCUTTED(argc, (*function), &context);
   }
@@ -848,7 +892,7 @@ callCleanupHandler(DECL_LD LocalFrame fr, enum finished reason)
       { int rval;
 
 	startCritical();
-        rval = call_term(contextModule(fr), clean);
+	rval = call_term(contextModule(fr), clean);
 	if ( !endCritical() )
 	  rval = FALSE;
 	if ( !rval && exception_term )
@@ -1289,6 +1333,8 @@ callBreakHook(DECL_LD LocalFrame frame, Choice bfr,
   wakeup_state wstate;
   size_t pc_offset;
 
+  SAVE_PTRS();
+
   *pop = 0;
   if (op == B_UNIFY_VAR || op == B_UNIFY_FIRSTVAR)
   { LD->slow_unify = TRUE;
@@ -1299,13 +1345,11 @@ callBreakHook(DECL_LD LocalFrame frame, Choice bfr,
   if ( !getProcDefinition(proc)->impl.any.defined )
     goto default_action;
 
-  if ( strchr(codeTable[op].argtype, CA1_FVAR) )
+  if ( strchr(VM_ARGTYPES(&codeTable[op]), CA1_FVAR) )
     pc_offset = stepPC(PC)-PC;
   else
     pc_offset = 0;
 
-  SAVE_PTRS();
-
   /* make enough space to avoid GC/shift in the	critical region*/
   if ( !hasGlobalSpace(10) )
   { int rc;
@@ -1319,19 +1363,21 @@ callBreakHook(DECL_LD LocalFrame frame, Choice bfr,
   if ( saveWakeup(&wstate, FALSE) )
   { if ( (cid=PL_open_foreign_frame()) )
     { term_t argv = PL_new_term_refs(7);
-      Clause clause = frame->clause->value.clause;
+      Clause clause;
       qid_t qid;
 
       RESTORE_PTRS();
+
+      clause = frame->clause->value.clause;
+      PL_put_frame(argv+2, frame); /* first, as it may move */
       PL_put_clref(argv+0, clause);
       PL_put_intptr(argv+1, PC - clause->codes);
-      PL_put_frame(argv+2, frame);
       PL_put_choice(argv+3, bfr);
       PL_put_bool(argv+5, debugstatus.debugging);
       if ( ( op == B_UNIFY_EXIT &&
-             put_call_goal(argv+4, GD->procedures.equals2) &&
-             PL_cons_functor_v(argv+4, FUNCTOR_call1, argv+4) ) ||
-           put_vm_call(argv+4, frameref, PC, op, pc_offset != 0, pop) )
+	     put_call_goal(argv+4, GD->procedures.equals2) &&
+	     PL_cons_functor_v(argv+4, FUNCTOR_call1, argv+4) ) ||
+	   put_vm_call(argv+4, frameref, PC, op, pc_offset != 0, pop) )
       { DEBUG(CHK_SECURE, checkStacks(NULL));
 	if ( (qid = PL_open_query(MODULE_user,
 				  PL_Q_NODEBUG|PL_Q_PASS_EXCEPTION, proc, argv)) )
@@ -1485,15 +1531,29 @@ __do_undo(DECL_LD mark *m)
       setVar(*p);
   }
 
+  DEBUG(CHK_SECURE,
+	{ for(tt = tTop; --tt >= mt;)
+	    tt->address = (Word)0xbfbfbfbf;
+	});
+
   tTop = mt;
+
+  Word ngtop;
   if ( LD->frozen_bar > m->globaltop )
   { DEBUG(CHK_SECURE, assert(gTop >= LD->frozen_bar));
     reclaim_attvars(LD->frozen_bar);
-    gTop = LD->frozen_bar;
+    ngtop = LD->frozen_bar;
   } else
   { reclaim_attvars(m->globaltop);
-    gTop = m->globaltop;
+    ngtop = m->globaltop;
   }
+
+  DEBUG(CHK_SECURE,
+	{ for(Word p = gTop; --p > ngtop;)
+	    *p = 0xbfbfbfbf;
+	});
+
+  gTop = ngtop;
 }
 
 
@@ -2625,7 +2685,7 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
   qf->saved_ltop = lTop;
 
 					/* fill top-frame */
-  top	             = &qf->top_frame;
+  top		     = &qf->top_frame;
   top->parent        = NULL;
   top->predicate     = PROCEDURE_dc_call_prolog->definition;
   top->programPointer= NULL;
@@ -2663,6 +2723,7 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
   qf->foreign_frame	= 0;
   qf->flags		= flags;
   qf->saved_environment = environment_frame;
+  assert(parentFrame(top) == environment_frame);
   qf->saved_bfr		= LD->choicepoints;
   qf->aSave             = aTop;
   qf->solutions         = 0;
@@ -3062,20 +3123,20 @@ typedef struct register_file
 #define _VMH_EPILOGUE			;
 
 /* Same syntax as VMH_GOTO, but handles profiling as if it were an instruction */
-#define VMH_GOTO_AS_VMI(n,args...) do { VMI_EXIT; \
-					VMI_ENTER(n); \
-					VMH_GOTO(n, args); \
-				      } while(0)
+#define VMH_GOTO_AS_VMI(n,...) do { VMI_EXIT; \
+				    VMI_ENTER(n); \
+				    VMH_GOTO(n, __VA_ARGS__); \
+				  } while(0)
 
 
 /* Helper macros for rendering VMH arguments */
 #define HEAD(h, ...) h
-#define TAIL(_, t...) (t)
-#define VMH_ARGS0(n,at,an,f,asep...)
-#define VMH_ARGS1(n,at,an,f,asep...) f(n, HEAD at, HEAD an)
-#define VMH_ARGS2(n,at,an,f,asep...) f(n, HEAD at, HEAD an) asep VMH_ARGS1(n, TAIL at, TAIL an, f, asep)
-#define VMH_ARGS3(n,at,an,f,asep...) f(n, HEAD at, HEAD an) asep VMH_ARGS2(n, TAIL at, TAIL an, f, asep)
-#define VMH_ARGS4(n,at,an,f,asep...) f(n, HEAD at, HEAD an) asep VMH_ARGS3(n, TAIL at, TAIL an, f, asep)
+#define TAIL(_, ...) (__VA_ARGS__)
+#define VMH_ARGS0(n,at,an,f,...)
+#define VMH_ARGS1(n,at,an,f,...) f(n, HEAD at, HEAD an)
+#define VMH_ARGS2(n,at,an,f,...) f(n, HEAD at, HEAD an) __VA_ARGS__ VMH_ARGS1(n, TAIL at, TAIL an, f, __VA_ARGS__)
+#define VMH_ARGS3(n,at,an,f,...) f(n, HEAD at, HEAD an) __VA_ARGS__ VMH_ARGS2(n, TAIL at, TAIL an, f, __VA_ARGS__)
+#define VMH_ARGS4(n,at,an,f,...) f(n, HEAD at, HEAD an) __VA_ARGS__ VMH_ARGS3(n, TAIL at, TAIL an, f, __VA_ARGS__)
 #define COMMA_TYPE_ARG(n,at,an) , at an
 #define TYPE_ARG_SEMI(n,at,an)	at an ;
 #define VMH_ARGS(n) A_PASTE(VMH_ARGS, VMH_ARGCOUNT(n))(n, (VMH_ARGTYPES(n)), (VMH_ARGNAMES(n)), TYPE_ARG_SEMI)
@@ -3083,8 +3144,18 @@ typedef struct register_file
 /* Define struct types for all the helper argument lists */
 #define VMH_ARGSTRUCT(Name)		struct helper_args_ ## Name
 
+/* GCC and CLang allow for struct name {}, i.e., an empty struct */
+#if O_EMPY_STRUCTS
+#define VMH_PAD_STRUCT
+#define VMH_INIT_ARGSTRUCT(...) {__VA_ARGS__}
+#else
+#undef HAVE_EMPTY_STRUCT
+#define VMH_PAD_STRUCT int _no_empty_struct;
+#define VMH_INIT_ARGSTRUCT(...) {0, ##__VA_ARGS__}
+#endif
+
 FOREACH_VMH(T_EMPTY,
-  ,VMH_ARGSTRUCT, { ,VMH_ARGS, };
+  ,VMH_ARGSTRUCT, {VMH_PAD_STRUCT ,VMH_ARGS, };
 )
 
 #define ASSIGN_ARG(n,at,an)		at an = HELPER_ARGS(n).an;
@@ -3098,7 +3169,7 @@ FOREACH_VMH(T_EMPTY,
 #define _NEXT_INSTRUCTION		return PC
 #define _SOLUTION_RETURN(val)		SOLUTION_RET = (val); longjmp(EXIT_VM_BUF, 1)
 #define _VMI_GOTO(n)			PC--; return instr_ ## n(VMI_ARG_PASS)
-#define _VMH_GOTO(n,...)		VMH_ARGSTRUCT(n) __args = {__VA_ARGS__}; (void)__args; \
+#define _VMH_GOTO(n,...)		VMH_ARGSTRUCT(n) __args = VMH_INIT_ARGSTRUCT(__VA_ARGS__); (void)__args; \
 					return helper_ ## n(VMI_ARG_PASS, __args)
 #undef _VMI_PROLOGUE
 #define _VMI_PROLOGUE(Ident,f,na,a)	PC++;
@@ -3148,7 +3219,8 @@ static vmi_instr jmp_table[] =
 
 #define HELPER_ARGS(n)			helper_args.n
 #define _VMH_DECLARATION(Name,na,at,an)	helper_ ## Name:
-#define _VMH_GOTO(n,args...)		VMH_ARGSTRUCT(n) __args = {args}; \
+#define _VHM_GOTO(n)			goto helper_ ## n;
+#define _VMH_GOTO(n,...)		VMH_ARGSTRUCT(n) __args = VMH_INIT_ARGSTRUCT(__VA_ARGS__); \
 					HELPER_ARGS(n) = __args; \
 					goto helper_ ## n;
 #define _SOLUTION_RETURN		return
@@ -3332,10 +3404,8 @@ variables used in the B_THROW instruction.
       DEBUG(CHK_SECURE, checkStacks(NULL));
       if ( exception_term )
 	THROW_EXCEPTION;
-      if ( QF->yield.term == -1 )		/* foreign yield */
-      { DEF = FR->predicate;
-	QF->yield.term = 0;
-      }
+      DEF = FR->predicate;
+      QF->yield.term = 0;
       NEXT_INSTRUCTION;
     } else
       BODY_FAILED;
